""" Multi-Head Attention module """
import math
import torch
import torch.nn as nn
from torch.autograd import Variable


class MultiHeadedAttention(nn.Module):
    """
    Multi-Head Attention module from
    "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`.
    Similar to standard `dot` attention but uses
    multiple attention distributions simulataneously
    to select relevant items.
    .. mermaid::
       graph BT
          A[key]
          B[value]
          C[query]
          O[output]
          subgraph Attn
            D[Attn 1]
            E[Attn 2]
            F[Attn N]
          end
          A --> D
          C --> D
          A --> E
          C --> E
          A --> F
          C --> F
          D --> O
          E --> O
          F --> O
          B --> O
    Also includes several additional tricks.
    Args:
       head_count (int): number of parallel heads
       model_dim (int): the dimension of keys/values/queries,
           must be divisible by head_count
       dropout (float): dropout parameter
    """

    def __init__(self, head_count, model_dim, d_k, d_v, dropout=0.1, clip_dist=0, use_neg_dist=False):
        super(MultiHeadedAttention, self).__init__()

        self.head_count = head_count
        self.model_dim = model_dim
        self.d_k = d_k
        self.d_v = d_v
        self.clip_dist = clip_dist
        self.use_neg_dist = use_neg_dist

        self.linear_keys = nn.Linear(model_dim, head_count * self.d_k)
        self.linear_query = nn.Linear(model_dim, head_count * self.d_k)
        self.linear_values = nn.Linear(model_dim, head_count * self.d_v)

        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.final_linear = nn.Linear(self.head_count * d_v, model_dim)

        # for relative-position aware self-attention
        if self.clip_dist > 0:
            self.edge_keys = nn.Parameter(torch.FloatTensor(2 * self.clip_dist + 1, self.d_k))
            self.edge_values = nn.Parameter(torch.FloatTensor(2 * self.clip_dist + 1, self.d_v))
            torch.nn.init.xavier_normal_(self.edge_keys)
            torch.nn.init.xavier_normal_(self.edge_values)

    def forward(self, key, value, query, mask=None):
        """
        Compute the context vector and the attention vectors.
        Args:
           key (`FloatTensor`): set of `key_len`
                key vectors `[batch, key_len, dim]`
           value (`FloatTensor`): set of `key_len`
                value vectors `[batch, key_len, dim]`
           query (`FloatTensor`): set of `query_len`
                 query vectors  `[batch, query_len, dim]`
           mask: binary mask indicating which keys have
                 non-zero attention `[batch, query_len, key_len]`
        Returns:
           (`FloatTensor`, `FloatTensor`) :
           * output context vectors `[batch, query_len, dim]`
           * one of the attention vectors `[batch, query_len, key_len]`
        """

        # CHECKS
        # batch, k_len, d = key.size()
        # batch_, k_len_, d_ = value.size()
        # aeq(batch, batch_)
        # aeq(k_len, k_len_)
        # aeq(d, d_)
        # batch_, q_len, d_ = query.size()
        # aeq(batch, batch_)
        # aeq(d, d_)
        # aeq(self.model_dim % 8, 0)
        # if mask is not None:
        #    batch_, q_len_, k_len_ = mask.size()
        #    aeq(batch_, batch)
        #    aeq(k_len_, k_len)
        #    aeq(q_len_ == q_len)
        # END CHECKS

        batch_size = key.size(0)
        head_count = self.head_count
        key_len = key.size(1)
        query_len = query.size(1)

        def shape(x, dim):
            """  projection """
            return x.view(batch_size, -1, head_count, dim).transpose(1, 2)

        def unshape(x, dim):
            """  compute context """
            return x.transpose(1, 2).contiguous().view(batch_size, -1, head_count * dim)

        if self.clip_dist > 0:
            dist_x = torch.arange(0, key_len).unsqueeze(0)
            dist_y = torch.arange(0, key_len).unsqueeze(1)
            distance = dist_x - dist_y
            if not self.use_neg_dist:
                distance = torch.abs(distance)
            distance = torch.clamp(distance, min=-self.clip_dist, max=self.clip_dist)
            distance = Variable((distance + self.clip_dist).long())     # [1,1,len,len]
            # distance = distance.repeat(head_count, 1, 1)  # nhead x seq_len x seq_len
            # distance = distance.repeat(batch_size, 1, 1, 1)  # bsz x nhead x seq_len x seq_len
            distance = distance.cuda() if key.is_cuda else distance

        # 1) Project key, value, and query.
        # key_up: bsz x nhead x key_len x d_k
        key_up = shape(self.linear_keys(key), self.d_k)  # x_j W^K
        # value_up: bsz x nhead x key_len x d_v
        value_up = shape(self.linear_values(value), self.d_v)  # x_j W^V
        # query_up: bsz x nhead x query_len x d_k
        query_up = shape(self.linear_query(query), self.d_k)  # x_j W^Q

        # 2) Calculate and scale scores.
        query_up = query_up / math.sqrt(self.d_k)  # (x_j W^Q) / sqrt(d_k)
        # bsz x nhead x query_len x key_len
        scores = torch.matmul(query_up, key_up.transpose(2, 3))
        if self.clip_dist > 0:
            out = self.edge_keys.index_select(0, distance.view(-1))
            # 1 x 1 x key_len x key_len x d_k
            # out = out.view(1, 1, query_len, query_len, self.d_k).contiguous()
            out = out.view(1, 1, query_len, query_len, self.d_k)
            # bsz x nhead x query_len x key_len
            add_term = torch.matmul(query_up.unsqueeze(3), out.transpose(3, 4)).squeeze(3)
            scores = scores + add_term

        if mask is not None:
            mask = mask.unsqueeze(1).expand_as(scores)
            scores = scores.masked_fill(mask, -1e20)

        # 3) Apply attention dropout and compute context vectors.
        attn = self.softmax(scores)
        drop_attn = self.dropout(attn)  # bsz x nhead x seq_len x seq_len
        context = torch.matmul(drop_attn, value_up)     # bsz x nhead x seq_len x d_v
        if self.clip_dist > 0:
            # TODO: modify `value_up` tensor to add a_{ij}^V [Eq.3 in https://arxiv.org/pdf/1803.02155.pdf]
            # also split Eq.3 to save space
            out = self.edge_values.index_select(0, distance.view(-1))
            # 1 x 1 x key_len x key_len x dim
            # out = out.view(1, 1, key_len, key_len, self.d_v).contiguous()
            out = out.view(1, 1, key_len, key_len, self.d_v)
            add_term = torch.matmul(drop_attn.unsqueeze(3), out).squeeze(3)     # bsz x nhead x seq_len x d_v
            context = context + add_term

        # bsz x nhead x seq_len x d_v --> bsz x seq_len x (nhead * d_v)
        context = unshape(context, self.d_v)

        # bsz x seq_len x (nhead * d_v) --> bsz x seq_len x model_dim
        output = self.final_linear(context)
        # CHECK
        # batch_, q_len_, d_ = output.size()
        # aeq(q_len, q_len_)
        # aeq(batch, batch_)
        # aeq(d, d_)

        # Return one attn
        top_attn = attn.view(batch_size, head_count, query_len, key_len)[:, 0, :, :].contiguous()

        return output, top_attn



#

# the multi-lingual version of create_alphabets

import os.path
import random
import numpy as np
from ..io.alphabet import Alphabet
from ..io.logger import get_logger
from ..io import utils

from .lang_id import guess_language_id, lang_specific_word

# Special vocabulary symbols - we always put them at the start.
PAD = b"_PAD"
PAD_POS = b"_PAD_POS"
PAD_TYPE = b"_<PAD>"
PAD_CHAR = b"_PAD_CHAR"
ROOT = b"_ROOT"
ROOT_POS = b"_ROOT_POS"
ROOT_TYPE = b"_<ROOT>"
ROOT_CHAR = b"_ROOT_CHAR"
END = b"_END"
END_POS = b"_END_POS"
END_TYPE = b"_<END>"
END_CHAR = b"_END_CHAR"
_START_VOCAB = [PAD, ROOT, END]

# a generator that returns the stream of (orig_tokens, normed_words, pos, types)
def iter_file(filename):
    with open(filename, 'r') as file:
        ret = {"len": 0, "word": [], "pos": [], "type": []}
        for line in file:
            line = line.strip()
            # yield and reset
            if len(line) == 0 or line[0] == "#":
                if ret["len"] > 0:
                    yield ret
                ret = {"len": 0, "word": [], "pos": [], "type": []}
            else:
                fields = line.split('\t')
                ret["len"] += 1
                ret["word"].append(fields[1])
                ret["pos"].append(fields[4])
                # ret["type"].append(fields[7])
                ret["type"].append(utils.get_main_deplabel(fields[7]))
        if ret["len"] > 0:
            yield ret

def create_alphabets(alphabet_directory, train_path, data_paths=None, max_vocabulary_size=100000, embedd_dict=None,
                     min_occurence=1, normalize_digits=True, creating_mode=False):
    logger = get_logger("Create Alphabets")
    word_alphabet = Alphabet('word', defualt_value=True, singleton=True)
    char_alphabet = Alphabet('character', defualt_value=True)
    pos_alphabet = Alphabet('pos')
    type_alphabet = Alphabet('type')
    max_sent_length = 0
    # guess language
    lang_id_train = guess_language_id(train_path)
    lang_id_extras = None if data_paths is None else [guess_language_id(fname) for fname in data_paths]
    logger.info("Here, the input files are: train(%s), extras(%s)." % (lang_id_train, lang_id_extras))
    if not os.path.isdir(alphabet_directory):
        logger.info("Creating Alphabets: %s" % alphabet_directory)
        # add special tokens
        char_alphabet.add(PAD_CHAR)
        pos_alphabet.add(PAD_POS)
        type_alphabet.add(PAD_TYPE)

        char_alphabet.add(ROOT_CHAR)
        pos_alphabet.add(ROOT_POS)
        type_alphabet.add(ROOT_TYPE)

        char_alphabet.add(END_CHAR)
        pos_alphabet.add(END_POS)
        type_alphabet.add(END_TYPE)

        # special one for Chinese
        type_alphabet.add("clf")

        # count from the main train file
        vocab = dict()
        for one_sent in iter_file(train_path):
            cur_len = one_sent["len"]
            max_sent_length = max(max_sent_length, cur_len)
            for idx in range(cur_len):
                cur_word, cur_pos, cur_type = one_sent["word"][idx], one_sent["pos"][idx], one_sent["type"][idx]
                #
                for char in cur_word:
                    char_alphabet.add(char)
                pos_alphabet.add(cur_pos)
                type_alphabet.add(cur_type)
                normed_word = utils.DIGIT_RE.sub(b"0", cur_word) if normalize_digits else cur_word
                # add prefix
                normed_word = lang_specific_word(normed_word, lang_id=lang_id_train)
                if normed_word in vocab:
                    vocab[normed_word] += 1
                else:
                    vocab[normed_word] = 1
        # collect singletons
        singletons = set([word for word, count in list(vocab.items()) if count <= min_occurence])
        # if a singleton is in pretrained embedding dict, set the count to min_occur + c
        if embedd_dict is not None:
            for word in list(vocab.keys()):
                if word in embedd_dict or word.lower() in embedd_dict:
                    vocab[word] += min_occurence
        #
        vocab_list = _START_VOCAB + sorted(vocab, key=vocab.get, reverse=True)
        logger.info("Total Vocabulary Size: %d" % len(vocab_list))
        logger.info("Total Singleton Size:  %d" % len(singletons))
        vocab_list = [word for word in vocab_list if word in _START_VOCAB or vocab[word] > min_occurence]
        logger.info("Total Vocabulary Size (w.o rare words): %d" % len(vocab_list))
        #
        if len(vocab_list) > max_vocabulary_size:
            vocab_list = vocab_list[:max_vocabulary_size]
        # extra directly added files (usually dev or test)
        if data_paths is not None:
            for one_path, one_lang_id in zip(data_paths, lang_id_extras):
                vocab_set = set(vocab_list)
                count_word_vocab_in_embed = 0
                count_word_all, count_word_in = 0, 0
                for one_sent in iter_file(one_path):
                    cur_len = one_sent["len"]
                    max_sent_length = max(max_sent_length, cur_len)
                    for idx in range(cur_len):
                        cur_word, cur_pos, cur_type = one_sent["word"][idx], one_sent["pos"][idx], one_sent["type"][idx]
                        #
                        for char in cur_word:
                            char_alphabet.add(char)
                        pos_alphabet.add(cur_pos)
                        type_alphabet.add(cur_type)
                        normed_word = utils.DIGIT_RE.sub(b"0", cur_word) if normalize_digits else cur_word
                        # add prefix
                        normed_word = lang_specific_word(normed_word, lang_id=one_lang_id)
                        if embedd_dict is not None:
                            if normed_word in embedd_dict or normed_word.lower() in embedd_dict:
                                if normed_word not in vocab_set:
                                    vocab_list.append(normed_word)
                                    count_word_vocab_in_embed += 1
                                count_word_in += 1
                        vocab_set.add(normed_word)
                        count_word_all += 1
                logger.info("For the file %s, vocab-size/in-size: %d/%d, word-size/in-size: %d/%d." % (one_path, len(vocab_set), count_word_vocab_in_embed, count_word_all, count_word_in))
        #
        for word in vocab_list:
            word_alphabet.add(word)
            if word in singletons:
                word_alphabet.add_singleton(word_alphabet.get_index(word))
        #
        word_alphabet.save(alphabet_directory)
        char_alphabet.save(alphabet_directory)
        pos_alphabet.save(alphabet_directory)
        type_alphabet.save(alphabet_directory)
    else:
        assert not creating_mode, "Cannot load existed vocabs in creating mode."
        logger.info("Loading Alphabets: %s" % alphabet_directory)
        word_alphabet.load(alphabet_directory)
        char_alphabet.load(alphabet_directory)
        pos_alphabet.load(alphabet_directory)
        type_alphabet.load(alphabet_directory)
    #
    word_alphabet.close()
    char_alphabet.close()
    pos_alphabet.close()
    type_alphabet.close()
    logger.info("Word Alphabet Size (Singleton): %d (%d)" % (word_alphabet.size(), word_alphabet.singleton_size()))
    logger.info("Character Alphabet Size: %d" % char_alphabet.size())
    logger.info("POS Alphabet Size: %d" % pos_alphabet.size())
    logger.info("Type Alphabet Size: %d" % type_alphabet.size())
    logger.info("Maximum Sentence Length: %d" % max_sent_length)
    #
    return word_alphabet, char_alphabet, pos_alphabet, type_alphabet, max_sent_length



import os

from typing import Union
from aksara.core import analyze_sentence
from aksara.analyzer import BaseAnalyzer
from dependency_parsing.core import DependencyParser


class Lemmatizer:
    def __init__(self):
        self.default_analyzer = self.__get_default_analyzer()
        self.default_dependency_parser = self.__get_default_dependency_parser()

    def lemmatize(self, word_input: str, is_informal: bool = False) -> Union[str, list]:
        """

        performs lemmatization on a word,

        then returns a string containing its corresponding

        lemma as the result or a list if the word contains

        certain affixes



        parameters
        ----------

        input: str

            the word that will be analyzed



        is_informal: bool

            tell aksara to treat text as informal , default to False



        return
        ------

        ReturnType: str | list

            will return string containing each word's lemma or
            list if the word contains certain affixes

        """
        word_input = word_input.strip()

        if word_input == "":
            return ""

        lemmatized_result = self.lemmatize_batch([word_input], is_informal)

        if len(lemmatized_result) > 1:
            return lemmatized_result

        result = lemmatized_result[0][1]

        return result

    def lemmatize_batch(
        self, list_word: list, is_informal: bool = False
    ) -> list[tuple[str, str]]:
        """

        performs lemmatization on the sentence list,

        then returns a list containing each word paired with its

        corresponding lemma as the result



        parameters
        ----------

        input: list

            the list of words (sentence) that will be analyzed



        is_informal: bool

            tell aksara to treat text as informal , default to False



        return
        ------

        ReturnType: list

            will return list containing each word paired with its

            corresponding lemma as the result

        """

        if list_word == []:
            return []

        input_text = " ".join(list_word)

        temp_result = analyze_sentence(
            text=input_text,
            analyzer=self.default_analyzer,
            dependency_parser=self.default_dependency_parser,
            v1=False,
            lemma=True,
            postag=False,
            informal=is_informal,
        )

        result = []

        for line in temp_result.split("\n"):
            _, word, lemma = line.split("\t")
            result.append((word, lemma))

        return result

    def __get_default_analyzer(self) -> BaseAnalyzer:
        bin_path = os.path.join(os.path.dirname(__file__), "bin", "aksara@v1.2.0.bin")
        return BaseAnalyzer(bin_path)

    def __get_default_dependency_parser(self) -> DependencyParser:
        return DependencyParser()



from typing import List
from igraph import Graph
import igraph
from aksara.conllu import ConlluData

def filter_multiword(conllu_datas: List[ConlluData]) -> List[ConlluData]:

    filtered_conllu = list(
        filter(lambda conllu_row: '-' not in conllu_row.get_id(), conllu_datas)
    )

    return filtered_conllu

def get_conllu_root(conllu_datas: List[ConlluData]) -> int:
    """get the index (start from 0) of the conllu root (row with head_id='0' or deprel='root')"""

    for idx, conllu_data in enumerate(conllu_datas):
        if conllu_data.get_head_id() == '0':
            return idx

    return -1

def create_igraph(conllu_datas: List[ConlluData]) -> Graph:
    edge_list = create_edge_list(conllu_datas)

    num_of_vertexs = len(conllu_datas)

    graph = Graph(n=num_of_vertexs, edges=edge_list)

    vertex_names = []
    for conllu_data in conllu_datas:
        vs_name = f"{conllu_data.get_form()}\n"

        if conllu_data.get_deprel() != '_':
            vs_name += f"{conllu_data.get_deprel()}\n"

        vs_name += conllu_data.get_upos()

        vertex_names.append(vs_name)

    graph.vs['name'] = vertex_names
    return graph

def create_edge_list(conllu_datas: List[ConlluData]) -> List[List[int]]:

    edge_list = []

    for my_idx, conllu_row in enumerate(conllu_datas):
        if conllu_row.get_head_id() in ['_', '0']:
            continue

        offset = int(conllu_row.get_id()) - my_idx
        parent_idx = int(conllu_row.get_head_id())
        edge_list.append([my_idx, parent_idx - offset])

    return edge_list

def is_in_ipython() -> bool:
    return hasattr(__builtins__, '__IPYTHON__')

def create_igraph_layout(
        graph: igraph.Graph,
        root: int,
        scale_x: float,
        scale_y: float
    ) -> igraph.Layout:

    # use Reingold-Tilford Tree layout
    layout = None
    if root == -1:
        layout = graph.layout('rt')
    else:
        layout = graph.layout('rt', root=[root])

    # put the root at the top of the tree
    layout.rotate(180)

    # rotate the graph with respect to y-axis
    # counteract the previuos rotation (for x-coordinate)
    layout.rotate(180, dim1=0, dim2=0)

    layout.scale([scale_x, scale_y])
    return layout



#!/usr/bin/python3

import re


class BaseTokenizer:
    _whitespace_pattern = r"\s+"
    _tokenize_pattern = r'([0-9]+\-an|[+-]?[0-9]*[,.]?[0-9]+|[A-Z][a-z]\.|(?:[A-Z]+\.)(?:[A-Za-z]+\.){1,}|[\w\-\.]+@(' \
                        r'[\w\-]+\.)+[\w\-]{2,4}|(?P<punct>[^\w\s+])(?P=punct)+|@[\w.]+|:[\S](?=\s|$)|:-[\S](' \
                        r'?=\s|$)|\w+(?=n\'t)|n\'t|\w+(?=\'[m|s]\s)|\'[m|s]\s|[^\w\s+]|(?:[\w-]{0,}))'

    def __init__(self):
        self.regex = re.compile(self._tokenize_pattern)
        self.whitespace_regex = re.compile(self._whitespace_pattern)

    def tokenize(self, sent):
        stripped_sent = self.whitespace_regex.sub(" ", sent).strip()
        tokens = self.regex.findall(stripped_sent)
        tokens = [token[0] for token in tokens]
        spaceafterflags = self.__getspaceafterflag(tokens)
        tokens = [token for token in tokens if token != '']
        return tokens, spaceafterflags

    def __getspaceafterflag(self, tokens):
        flag = [False for token in tokens if token != '']

        # Iterate over list
        i = 0
        is_whitespace = True
        for token in tokens:
            if token == '':
                is_whitespace = True
            else:
                flag[i] = not is_whitespace
                i += 1
                is_whitespace = False

        return flag[1:] + [False]



import unittest
from src.aksara.lemmatizer import Lemmatizer


class LemmatizationTest(unittest.TestCase):
    """This class contains unit testcases for the lemmatization feature"""

    def setUp(self) -> None:
        self.lemmatizer = Lemmatizer()
        return super().setUp()

    # Sentence lemmatization
    def test_sentence_lemmatization(self):
        testcase = ["saya", "batuk"]
        expected = [("saya", "saya"), ("batuk", "batuk")]

        self.assertEqual(self.lemmatizer.lemmatize_batch(testcase), expected)

    def test_sentence_lemmatization_informal(self):
        testcase = ["lg", "udh", "ngajakin"]
        expected = [("lg", "lagi"), ("udh", "sudah"), ("ngajakin", "ajak")]

        self.assertEqual(self.lemmatizer.lemmatize_batch(testcase, True), expected)

    def test_sentence_lemmatization_informal_false(self):
        testcase = ["lg", "udh", "ngajakin"]
        expected = [("lg", "lg"), ("udh", "udh"), ("ngajakin", "ngajakin")]

        self.assertEqual(self.lemmatizer.lemmatize_batch(testcase, False), expected)

    def test_affix_sentence_lemmatization(self):
        testcase = ["airnya", "menguning"]
        expected = [
            ("airnya", "_"),
            ("air", "air"),
            ("nya", "nya"),
            ("menguning", "kuning"),
        ]

        self.assertEqual(self.lemmatizer.lemmatize_batch(testcase), expected)

    def test_empty_sentence_lemmatization(self):
        testcase = []
        expected = []

        self.assertEqual(self.lemmatizer.lemmatize_batch(testcase, True), expected)

    def test_lemmatization_empty_string(self):
        testcase = ["belajar", ""]
        expected = [("belajar", "ajar")]

        self.assertEqual(self.lemmatizer.lemmatize_batch(testcase, True), expected)

    # Word lemmatization
    def test_word_lemmatization_with_affix(self):
        testcase = "Pengeluaran"
        expected = "keluar"

        self.assertEqual(self.lemmatizer.lemmatize(testcase), expected)

    def test_word_lemmatization_without_affix(self):
        testcase = "gemuk"
        expected = "gemuk"

        self.assertEqual(self.lemmatizer.lemmatize(testcase), expected)

    def test_symbol_lemmatization(self):
        testcase = "."
        expected = "."

        self.assertEqual(self.lemmatizer.lemmatize(testcase), expected)

    def test_word_lemmatization_whitespace(self):
        testcase = ""
        expected = ""

        self.assertEqual(self.lemmatizer.lemmatize(testcase), expected)

    def test_certain_affix_word_lemmatization(self):
        testcase = "Airnya"
        expected = [("Airnya", "_"), ("Air", "air"), ("nya", "nya")]

        self.assertEqual(self.lemmatizer.lemmatize(testcase), expected)



#!/usr/bin/python3

def to_conllu_line_with_range(line_id, surface, n_tokens):
    new_row = ['_'] * 10
    range_id = "{}-{}".format(line_id, line_id + n_tokens - 1)

    # basic info
    new_row[0] = range_id
    new_row[1] = surface

    return '\t'.join(new_row)


def to_conllu_line(line_id, surface, text, **kwargs):
    new_row = ['_'] * 10

    # Split classes
    candidates = text.split("\\n")
    candidates = [cand.split("+") for cand in candidates]

    # Disambiguation, if any (further project)
    appended_lemma = ""
    appended_tags = ""
    appended_features = ""
    ambiguous_features_template = "({} -> {})"
    appended_misc = ""

    for cand in candidates:
        appended_lemma += cand[0] + "/"
        appended_tags += cand[1] + "/"
        if len(cand) > 2:
            # making sure the features are sorted
            misc = "Morf="
            feats_list = sorted([element.split("=") for element in cand[2:]])
            prepref = [element[1] for element in feats_list if element[0] == "Prepref"]
            pref = [element[1] for element in feats_list if element[0] == "Pref"]
            stem = [element[1] for element in feats_list if element[0] == "Stem"]
            suff = [element[1] for element in feats_list if element[0] == "Suff"]
            if prepref != []:
                misc += "".join(prepref)
                misc += "+"
            if pref != []:
                misc += "".join(pref)
                misc += "+"
            misc += cand[0]
            if stem != []:
                misc += "<" + "".join(stem) + ">"
            elif cand[1] != 'X':
                misc += "<" + cand[1] + ">"
            else:
                misc += "<X>"
            if suff != []:
                misc += "+"
                misc += "".join(suff)
            misc += "_" + cand[1]

            for key, value in kwargs.items():
                if key == 'space_after' and value:
                    misc += "|SpaceAfter=No"
            if misc == "":
                misc = "_"
            appended_misc += misc + "/"

            feats_string = "|".join(
                "=".join(element) for element in feats_list if element[0] not in ["Prepref", "Pref", "Stem", "Suff"])
            if feats_string == "":
                feats_string = "_"
            if len(candidates) > 1:
                feats_string = ambiguous_features_template.format(cand[1], feats_string)
            appended_features += feats_string + "/"
        else:
            misc = "Morf="
            misc += cand[0]
            if cand[1] != 'X':
                misc += "<" + cand[1] + ">"
            else:
                misc += "<X>"
            misc += "_" + cand[1]
            appended_misc += misc + "/"

    if appended_features == "" or appended_features == "/":
        appended_features = "_/"

    appended_lemma = appended_lemma[:-1]
    appended_tags = appended_tags[:-1]
    appended_features = appended_features[:-1]
    appended_misc = appended_misc[:-1]

    candidate = [appended_lemma, appended_tags, appended_features]

    # basic info
    new_row[0] = str(line_id)
    new_row[1] = surface

    # lemma
    new_row[2] = candidate[0]

    # surface POS-tag
    new_row[3] = candidate[1]

    # lexical feature
    new_row[5] = candidate[2]

    # misc
    new_row[9] = appended_misc

    return '\t'.join(new_row)



from typing import List, Type, Any
from aksara.conllu import ConlluData
from ._drawer import MatplotlibDrawer, AbstractDrawer, PlotlyDrawer

class TreeDrawer:
    """
    Draw dependency tree with specific drawer (e.g. Matplotlib, Plotly, etc)

    Attributes
    ----------
    name_to_drawer: dict
        mapping drawer name with its instance
    
    Methods
    -------
    draw(conllu_datas)
        Draw dependency tree based on `conllu_datas`
    
    save_image(conllu_datas, ouput_path)
        Save dependency tree image

    """

    name_to_drawer = {
        'matplotlib' : MatplotlibDrawer(),
        'plotly': PlotlyDrawer()
    }

    def __init__(self, drawer: str = 'matplotlib') -> None:
        self.set_drawer(drawer)

    def draw(
        self,
        conllu_datas: List[List[ConlluData]],
        *args,
        **kwargs
    ) -> Any:
        """
        Draw dependency tree

        Parameters
        ----------
        conllu_datas: list of list of ConlluData
            conllu_datas that will be drawn
        
        *args: tuple
            will be passed to its drawer
        
        **kwargs
            will be passed to its drawer

        Returns
        -------
        Any
            See drawer return value

        See Also
        --------
        aksara.dependency_tree._drawer.MatplotlibDrawer: Matplotlib drawer instance
        aksara.dependency_tree._drawer.PlotlyDrawer: Plotly drawer instance
        """

        return self.__drawer.draw(conllu_datas, *args, **kwargs)

    def save_image(
        self,
        conllu_datas: List[List[ConlluData]],
        output_path: str,
        *args,
        **kwargs
    ) -> None:
        """
        Draw dependency tree

        Parameters
        ----------
        conllu_datas: list of list of ConlluData
            conllu_datas that will be drawn and saved
        
        output_path: str
            file path at which the dependency tree image will be saved
        
        *args: tuple
            will be passed to its drawer
        
        **kwargs
            will be passed to its drawer

        Returns
        -------
        None

        See Also
        --------
        aksara.dependency_tree._drawer.MatplotlibDrawer: Matplotlib drawer instance
        aksara.dependency_tree._drawer.PlotlyDrawer: Plotly drawer instance
        """

        self.__drawer.save_image(conllu_datas, output_path, *args, **kwargs)

    def get_drawer(self) -> Type[AbstractDrawer]:
        """drawer instance getter
        """

        return self.__drawer

    def set_drawer(self, drawer: str) -> None:
        """drawer setter
        """

        drawer_instance = TreeDrawer.name_to_drawer.get(drawer, None)

        if drawer_instance is None:
            raise ValueError(f"drawer name must be one of {TreeDrawer.name_to_drawer.keys()}")

        self.__drawer = drawer_instance



import os

# todo(warn): guess language id from filenames
# the ids should be lowercased!!
KNOWN_LANG_IDS = {'ar', 'bg', 'ca', 'zh', 'hr', 'cs', 'da', 'nl', 'en', 'et', 'fi', 'fr', 'de', 'he', 'hi', 'id', 'it', 'ja', 'ko', 'la', 'lv', 'no', 'pl', 'pt', 'ro', 'ru', 'sk', 'sl', 'es', 'sv', 'uk'}
def guess_language_id(file_path):
    bname = os.path.basename(file_path)
    lang_id = str.lower(bname[:2])
    # assert lang_id in KNOWN_LANG_IDS, "Unknown lang id %s from path %s" % (lang_id, file_path)
    if lang_id not in KNOWN_LANG_IDS:
        print(("Warning: Unknown lang id %s from path %s" % (lang_id, file_path)))
    return lang_id

def lang_specific_word(word, lang_id):
    if lang_id:
        # assert lang_id in KNOWN_LANG_IDS, "Unknown lang id %s" % (lang_id, )
        return "!%s_%s" % (lang_id, word)
    else:
        return word

# backoff to DEFAULT_LANG
DEFAULT_LANG_ID = "en"
ALREADY_PREFIX = "!en_"

def default_lang_specific_word(word):
    return lang_specific_word(word, DEFAULT_LANG_ID)

def get_word_index_with_spec(alphabet, word, lang_id):
    # if already prefixed (maybe by outside pre-processor, then go with it)
    if word.startswith(ALREADY_PREFIX):
        return alphabet.get_index(word)
    #
    prefixed_word = lang_specific_word(word, lang_id=lang_id)
    if prefixed_word in alphabet.instance2index:
        return alphabet.get_index(prefixed_word)
    else:
        # try to get the English(default) word's index and later its embedding
        prefixed_word = default_lang_specific_word(word)
        return alphabet.get_index(prefixed_word)



import collections
from itertools import repeat
import torch
import torch.nn.utils.rnn as rnn_utils
from torch.autograd import Variable


def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse

_single = _ntuple(1)
_pair = _ntuple(2)
_triple = _ntuple(3)
_quadruple = _ntuple(4)


def prepare_rnn_seq(rnn_input, lengths, hx=None, masks=None, batch_first=False):
    '''

    Args:
        rnn_input: [seq_len, batch, input_size]: tensor containing the features of the input sequence.
        lengths: [batch]: tensor containing the lengthes of the input sequence
        hx: [num_layers * num_directions, batch, hidden_size]: tensor containing the initial hidden state for each element in the batch.
        masks: [seq_len, batch]: tensor containing the mask for each element in the batch.
        batch_first: If True, then the input and output tensors are provided as [batch, seq_len, feature].

    Returns:

    '''
    def check_decreasing(lengths):
        lens, order = torch.sort(lengths, dim=0, descending=True)
        if torch.ne(lens, lengths).sum() == 0:
            return None
        else:
            _, rev_order = torch.sort(order)
            return lens, Variable(order), Variable(rev_order)

    check_res = check_decreasing(lengths)

    if check_res is None:
        lens = lengths
        rev_order = None
    else:
        lens, order, rev_order = check_res
        batch_dim = 0 if batch_first else 1
        rnn_input = rnn_input.index_select(batch_dim, order)
        if hx is not None:
            # hack lstm
            if isinstance(hx, tuple):
                hx, cx = hx
                hx = hx.index_select(1, order)
                cx = cx.index_select(1, order)
                hx = (hx, cx)
            else:
                hx = hx.index_select(1, order)

    lens = lens.tolist()
    seq = rnn_utils.pack_padded_sequence(rnn_input, lens, batch_first=batch_first)
    if masks is not None:
        if batch_first:
            masks = masks[:, :lens[0]]
        else:
            masks = masks[:lens[0]]
    return seq, hx, rev_order, masks


def recover_rnn_seq(seq, rev_order, hx=None, batch_first=False):
    output, _ = rnn_utils.pad_packed_sequence(seq, batch_first=batch_first)
    if rev_order is not None:
        batch_dim = 0 if batch_first else 1
        output = output.index_select(batch_dim, rev_order)
        if hx is not None:
            # hack lstm
            if isinstance(hx, tuple):
                hx, cx = hx
                hx = hx.index_select(1, rev_order)
                cx = cx.index_select(1, rev_order)
                hx = (hx, cx)
            else:
                hx = hx.index_select(1, rev_order)
    return output, hx



__author__ = 'max'


class Sentence(object):
    def __init__(self, words, word_ids, char_seqs, char_id_seqs):
        self.words = words
        self.word_ids = word_ids
        self.char_seqs = char_seqs
        self.char_id_seqs = char_id_seqs

    def length(self):
        return len(self.words)


class DependencyInstance(object):
    def __init__(self, sentence, postags, pos_ids, heads, types, type_ids):
        self.sentence = sentence
        self.postags = postags
        self.pos_ids = pos_ids
        self.heads = heads
        self.types = types
        self.type_ids = type_ids

    def length(self):
        return self.sentence.length()


class NERInstance(object):
    def __init__(self, sentence, postags, pos_ids, chunk_tags, chunk_ids, ner_tags, ner_ids):
        self.sentence = sentence
        self.postags = postags
        self.pos_ids = pos_ids
        self.chunk_tags = chunk_tags
        self.chunk_ids = chunk_ids
        self.ner_tags = ner_tags
        self.ner_ids = ner_ids

    def length(self):
        return self.sentence.length()


__author__ = 'max'

"""
Alphabet maps objects to integer ids. It provides two way mapping from the index to the objects.
"""
import json
import os
from .logger import get_logger

class Alphabet(object):
    def __init__(self, name, defualt_value=False, keep_growing=True, singleton=False):
        self.__name = name

        self.instance2index = {}
        self.instances = []
        self.default_value = defualt_value
        self.offset = 1 if self.default_value else 0
        self.keep_growing = keep_growing
        self.singletons = set() if singleton else None

        # Index 0 is occupied by default, all else following.
        self.default_index = 0 if self.default_value else None

        self.next_index = self.offset

        self.logger = get_logger('Alphabet')

    def add(self, instance):
        if instance not in self.instance2index:
            self.instances.append(instance)
            self.instance2index[instance] = self.next_index
            self.next_index += 1

    def add_singleton(self, id):
        if self.singletons is None:
            raise RuntimeError('Alphabet %s does not have singleton.' % self.__name)
        else:
            self.singletons.add(id)

    def add_singletons(self, ids):
        if self.singletons is None:
            raise RuntimeError('Alphabet %s does not have singleton.' % self.__name)
        else:
            self.singletons.update(ids)

    def is_singleton(self, id):
        if self.singletons is None:
            raise RuntimeError('Alphabet %s does not have singleton.' % self.__name)
        else:
            return id in self.singletons

    def get_index(self, instance):
        try:
            return self.instance2index[instance]
        except KeyError:
            if self.keep_growing:
                index = self.next_index
                self.add(instance)
                return index
            else:
                if self.default_value:
                    return self.default_index
                else:
                    raise KeyError("instance not found: %s" % instance)

    def get_instance(self, index):
        if self.default_value and index == self.default_index:
            # First index is occupied by the wildcard element.
            return '<_UNK>'
        else:
            try:
                return self.instances[index - self.offset]
            except IndexError:
                raise IndexError('unknown index: %d' % index)

    def size(self):
        return len(self.instances) + self.offset

    def singleton_size(self):
        return len(self.singletons)

    def items(self):
        return list(self.instance2index.items())

    def enumerate_items(self, start):
        if start < self.offset or start >= self.size():
            raise IndexError("Enumerate is allowed between [%d : size of the alphabet)" % self.offset)
        return list(zip(list(range(start, len(self.instances) + self.offset)), self.instances[start - self.offset:]))

    def close(self):
        self.keep_growing = False

    def open(self):
        self.keep_growing = True

    def get_content(self):
        if self.singletons is None:
            return {'instance2index': self.instance2index, 'instances': self.instances}
        else:
            return {'instance2index': self.instance2index, 'instances': self.instances,
                    'singletions': list(self.singletons)}

    def __from_json(self, data):
        self.instances = data["instances"]
        self.instance2index = data["instance2index"]
        if 'singletions' in data:
            self.singletons = set(data['singletions'])
        else:
            self.singletons = None

    def save(self, output_directory, name=None):
        """
        Save both alhpabet records to the given directory.
        :param output_directory: Directory to save model and weights.
        :param name: The alphabet saving name, optional.
        :return:
        """
        saving_name = name if name else self.__name
        try:
            if not os.path.exists(output_directory):
                os.makedirs(output_directory)

            json.dump(self.get_content(),
                      open(os.path.join(output_directory, saving_name + ".json"), 'w'), indent=4)
        except Exception as e:
            self.logger.warn("Alphabet is not saved: %s" % repr(e))

    def load(self, input_directory, name=None):
        """
        Load model architecture and weights from the give directory. This allow we use old models even the structure
        changes.
        :param input_directory: Directory to save model and weights
        :return:
        """
        loading_name = name if name else self.__name
        self.__from_json(json.load(open(os.path.join(input_directory, loading_name + ".json"))))
        self.next_index = len(self.instances) + self.offset
        self.keep_growing = False
